#!/usr/bin/env python
# -*- coding: utf-8 -*-

from copy import copy
from typing import Dict, Any, Optional

from .base import Block


class Button(Block):
    """
    Render a ``<button>`` with Bootstrap styling.  Example::

        from wildwidgets import Button

        Button(text="My Button")

    When rendered in the template with the ``wildewdigets`` template tag, this will produce::

        <button type="button" class="button btn btn-secondary">My Button</button>

    All the constructor parameters can be set in a subclass of this class as class attributes.  Parameters
    to the constructor override any defined class attributes.

    :param text: The text to use for the button, defaults to 'Button'
    :type text: str
    :param color: The Boostrap color class to use for the button, defaults to 'secondary'
    :type color: str
    :param close: The Boostrap close icon will be used for the button, defaults to False.
    :type close: bool
    :param name: This CSS class will be added to the classes to identify this button, defaults to 'block'
    :type tag: str
    :param modifier: If specified, also add a class named ``{name}--{modifier}`` to the CSS classes, defaults
        to no modifier
    :type modifier: str
    :param css_class: a string of classes to apply to the button, defaults to no classes.
    :type css_class: str
    :param css_id: Use this as the ``id`` attribute for the button, defaults to no ``id``
    :type css_id: str
    :param attributes: Set any additional attributes for the button as key, value pairs, defaults to no additional
        attributes.
    :type attributes: dict(str, str)
    :param data_attributes: Set ``data-`` attributes for the button, defaults to no data attributes
    :type data_attributes: dict(str, str)
    """
    tag: str = 'button'
    block: str = 'button'
    color: str = 'secondary'
    text: str = 'Button'

    def __init__(self, **kwargs):
        color = kwargs.pop('color', self.color)
        text = kwargs.pop('text', self.text)
        close = kwargs.pop('close', False)
        if close:
            text = ""
            kwargs["css_class"] = "btn-close"
            kwargs["aria_attributes"] = {"label": "Close"}
        super().__init__(text, **kwargs)
        self._content = text
        self._attributes['type'] = 'button'
        if not close:
            if self._css_class:
                self._css_class = f'{self._css_class} btn btn-{color}'
            else:
                self._css_class = f'btn btn-{color}'


class ModalButton(Button):
    """
    Render a ``<button>`` with Bootstrap styling which toggles a Bootstrap modal.  Example::

        from wildwidgets import ModalButton

        ModalButton(text="My Button", target='#mymodal')

    When rendered in the template with the ``wildewdigets`` template tag, this will produce::

        <button type="button" class="button button--modal btn btn-secondary" data-toggle="modal"
          data-target="#mymodal">My Button</button>

    All the constructor parameters can be set in a subclass of this class as class attributes.  Parameters
    to the constructor override any defined class attributes.

    :param text: The text to use for the button, defaults to 'Button'
    :type text: str
    :param color: The Boostrap color class to use for the button, defaults to 'secondary'
    :type color: str
    :param target: The CSS target for the Bootstrap modal, defaults to no target.
    :type target: str
    :param name: This CSS class will be added to the classes to identify this button, defaults to 'block'
    :type tag: str
    :param modifier: If specified, also add a class named ``{name}--{modifier}`` to the CSS classes, defaults
        to no modifier
    :type modifier: str
    :param css_class: a string of classes to apply to the button, defaults to no classes.
    :type css_class: str
    :param css_id: Use this as the ``id`` attribute for the button, defaults to no ``id``
    :type css_id: str
    :param attributes: Set any additional attributes for the button as key, value pairs, defaults to no additional
        attributes.
    :type attributes: dict(str, str)
    :param data_attributes: Set ``data-`` attributes for the button, defaults to no data attributes
    :type data_attributes: dict(str, str)
    """
    block: str = 'button button--modal'
    target: Optional[str] = None

    def __init__(self, **kwargs):
        target = kwargs.pop('target', self.target)
        super().__init__(**kwargs)
        self._data_attributes['toggle'] = 'modal'
        self._data_attributes['target'] = target if target is not None else self.target


class CollapseButton(Button):
    """
    Render a ``<button>`` with Bootstrap styling which toggles a Bootstrap modal.  Example::

        from wildwidgets import ModalButton

        ModalButton(text="My Button", target='#mymodal')

    When rendered in the template with the ``wildewdigets`` template tag, this will produce::

        <button type="button" class="button button--modal btn btn-secondary" data-toggle="modal"
          data-target="#mymodal">My Button</button>

    All the constructor parameters can be set in a subclass of this class as class attributes.  Parameters
    to the constructor override any defined class attributes.

    :param text: The text to use for the button, defaults to 'Button'
    :type text: str
    :param color: The Boostrap color class to use for the button, defaults to 'secondary'
    :type color: str
    :param target: The CSS target for the Bootstrap modal, defaults to no target.
    :type target: str
    :param name: This CSS class will be added to the classes to identify this button, defaults to 'block'
    :type tag: str
    :param modifier: If specified, also add a class named ``{name}--{modifier}`` to the CSS classes, defaults
        to no modifier
    :type modifier: str
    :param css_class: a string of classes to apply to the button, defaults to no classes.
    :type css_class: str
    :param css_id: Use this as the ``id`` attribute for the button, defaults to no ``id``
    :type css_id: str
    :param attributes: Set any additional attributes for the button as key, value pairs, defaults to no additional
        attributes.
    :type attributes: dict(str, str)
    :param data_attributes: Set ``data-`` attributes for the button, defaults to no data attributes
    :type data_attributes: dict(str, str)
    """
    block: str = 'button button--collapse'
    target: Optional[str] = None

    def __init__(self, **kwargs):
        target = kwargs.pop('target', self.target)
        super().__init__(**kwargs)
        self._data_attributes['toggle'] = 'collapse'
        self._data_attributes['target'] = target
        self._aria_attributes['expanded'] = 'false'
        self._aria_attributes['controls'] = target.lstrip("#")


class LinkButton(Button):
    """
    Render an ``<a>`` with Bootstrap button styling which toggles a Bootstrap modal.  Example::

        from wildwidgets import LinkButton

        LinkButton(text="My Button", url='https://myexample.com')

    When rendered in the template with the ``wildewdigets`` template tag, this will produce::

        <a href="https://myexample.com" class="button button--link btn btn-secondary">My Button</a>

    All the constructor parameters can be set in a subclass of this class as class attributes.  Parameters
    to the constructor override any defined class attributes.

    :param text: The text to use for the button, defaults to 'Button'
    :type text: str
    :param color: The Boostrap color class to use for the button, defaults to 'secondary'
    :type color: str
    :param url: The URL for the ``href`` attribute of the <a> record, defaults to no URL.
    :type target: str
    :param name: This CSS class will be added to the classes to identify this button, defaults to 'block'
    :type tag: str
    :param modifier: If specified, also add a class named ``{name}--{modifier}`` to the CSS classes, defaults
        to no modifier
    :type modifier: str
    :param css_class: a string of classes to apply to the button, defaults to no classes.
    :type css_class: str
    :param css_id: Use this as the ``id`` attribute for the button, defaults to no ``id``
    :type css_id: str
    :param attributes: Set any additional attributes for the button as key, value pairs, defaults to no additional
        attributes.
    :type attributes: dict(str, str)
    :param data_attributes: Set ``data-`` attributes for the button, defaults to no data attributes
    :type data_attributes: dict(str, str)
    """
    tag: str = 'a'
    block: str = 'button button--link'
    url: Optional[str] = None

    def __init__(self, **kwargs):
        url = kwargs.pop('url', self.url)
        super().__init__(**kwargs)
        del self._attributes['type']
        if url:
            self._attributes['href'] = url


class InputButton(Button):
    """
    Render an ``<input type="submit">`` with Bootstrap button styling.  Example::

        InputButton(text="Save")

    When rendered in the template with the ``wildewdigets`` template tag, this will produce::

        <input type="submit" class="button button--submit btn btn-secondary" value="Save">

    All the constructor parameters can be set in a subclass of this class as class attributes.  Parameters
    to the constructor override any defined class attributes.

    :param text: The text to use for the button, defaults to 'Button'
    :type text: str
    :param color: The Boostrap color class to use for the button, defaults to 'secondary'
    :type color: str
    :param url: The URL for the ``href`` attribute of the <a> record, defaults to no URL.
    :type target: str
    :param name: This CSS class will be added to the classes to identify this button, defaults to 'block'
    :type tag: str
    :param modifier: If specified, also add a class named ``{name}--{modifier}`` to the CSS classes, defaults
        to no modifier
    :type modifier: str
    :param css_class: a string of classes to apply to the button, defaults to no classes.
    :type css_class: str
    :param css_id: Use this as the ``id`` attribute for the button, defaults to no ``id``
    :type css_id: str
    :param attributes: Set any additional attributes for the button as key, value pairs, defaults to no additional
        attributes.
    :type attributes: dict(str, str)
    :param data_attributes: Set ``data-`` attributes for the button, defaults to no data attributes
    :type data_attributes: dict(str, str)
    """
    template_name: str = 'wildewidgets/block--simple.html'
    tag: str = 'input'
    block: str = 'button--submit'

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._attributes['type'] = 'submit'
        self._attributes['value'] = self._content
        self._content = None


class FormButton(Block):
    """
    Render a ``<form>`` with optional hidden inputs and a submit button.

        from wildwidgets import FormButton

        FormButton(text="Save", action='/my/form/action', data={'field1': 'value1'})

    When rendered in the template with the ``wildewdigets`` template tag, this will produce::

        <form action="/my/form/action" method="post" class="button-form">
            <input type="hidden" name="csrfmiddlewaretoken" value="__THE_CSRF_TOKEN__">
            <input type="hidden" name="field1" value="value1">
            <input type="submit" class="button button--submit btn btn-secondary" value="Save">
        </form>

    All the constructor parameters can be set in a subclass of this class as class attributes.  Parameters
    to the constructor override any defined class attributes.

    :param text: The text to use for the button, defaults to 'Button'
    :type text: str
    :param color: The Boostrap color class to use for the button, defaults to 'secondary'
    :type color: str
    :param url: The URL for the ``href`` attribute of the <a> record, defaults to no URL.
    :type target: str
    :param name: This CSS class will be added to the classes to identify this form, defaults to 'button-form'
    :type tag: str
    :param modifier: If specified, also add a class named ``{name}--{modifier}`` to the form CSS classes, defaults
        to no modifier
    :type modifier: str
    :param css_class: a string of classes to apply to the form, defaults to no classes.
    :type css_class: str
    :param button_css_class: a string of classes to apply to the button, defaults to no classes.
    :type button_css_class: str
    :param css_id: Use this as the ``id`` attribute for the form, defaults to no ``id``
    :type css_id: str
    :param attributes: Set any additional attributes for the form as key, value pairs, defaults to no additional
        attributes.
    :type attributes: dict(str, str)
    :param button_attributes: Set any additional attributes for the button as key, value pairs, defaults to no
        additional attributes.
    :type attributes: dict(str, str)
    :param data_attributes: Set ``data-`` attributes for the form, defaults to no data attributes
    :type data_attributes: dict(str, str)
    :param button_data_attributes: Set ``data-`` attributes for the button, defaults to no data attributes
    :type button_data_attributes: dict(str, str)
    """
    template_name = 'wildewidgets/button--form.html'
    tag: str = 'form'
    block: str = 'button-form'
    action: Optional[str] = None
    method: str = 'post'
    text: Optional[str] = None
    color: str = 'secondary'
    button_css_class: Optional[str] = None
    button_attributes: Dict[str, str] = {}
    button_data_attributes: Dict[str, str] = {}
    confirm_text: Optional[str] = None
    data: Dict[str, Any] = {}

    def __init__(self, **kwargs):
        action = kwargs.pop('action', self.action)
        method = kwargs.pop('method', self.method)
        self.close = kwargs.pop('close', False)
        self.data = kwargs.pop('data', copy(self.data))
        self.button = self.__build_button(kwargs)
        super().__init__(**kwargs)
        self._attributes['method'] = method
        self._attributes['action'] = action

    def __build_button(self, kwargs) -> InputButton:
        button_kwargs = {
            'text': kwargs.pop('text', self.text),
            'color': kwargs.pop('color', self.color),
            'css_class': kwargs.pop('button_css_class', self.button_css_class),
            'attributes': kwargs.pop('button_attributes', copy(self.button_attributes)),
            'data_attributes': kwargs.pop('button_data_attributes', copy(self.button_data_attributes))
        }
        if self.close:
            button_kwargs['close'] = True
            button_kwargs['aria_attributes'] = {"label": "Close"}
        confirm_text = kwargs.pop('confirm_text', self.confirm_text)
        if confirm_text is not None:
            button_kwargs['attributes']['onclick'] = f"return confirm('{confirm_text}');"
        return InputButton(**button_kwargs)

    def get_context_data(self, **kwargs) -> Dict[str, Any]:
        context = super().get_context_data(**kwargs)
        context['button'] = self.button
        context['data'] = self.data
        return context


class ButtonRow(Block):

    def __init__(self, *blocks, **kwargs):
        css_class = kwargs.get("css_class", "")
        css_class += " d-flex justify-content-end"
        kwargs["css_class"] = css_class
        super().__init__(*blocks, **kwargs)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from typing import Optional
import warnings

from .base import TemplateWidget, Block
from .buttons import ModalButton, CollapseButton, LinkButton, FormButton


class BasicHeader(TemplateWidget):
    """
    Extends :py:class:`wildewidgets.widgets.base.TemplateWidget`.

    The base header class, which contains the bulk of the functionality.

    Keyword Args:
        header_level: the header depth, as in h1, h2, h3...
        header_type: the Boostrap 5 header class, either h-*, or display-*.
        header_text: the text of the header.
        css_class: any css classes to apply to the header.
        css_id: an ID to add to the header div.
        badge_text: text to add in a badge to the right of the main header text.
        badge_class: the css class to add to the badge.
    """
    template_name: str = 'wildewidgets/header_with_controls.html'

    header_level: int = 1
    header_type: str = 'h'
    header_text: Optional[str] = None
    css_class: str = "my-3"
    css_id: Optional[str] = None
    badge_text: Optional[str] = None
    badge_class: str = "warning"

    def __init__(
        self,
        header_level: int = None,
        header_type: str = None,
        header_text: str = None,
        css_class: str = None,
        css_id: str = None,
        badge_text: str = None,
        badge_class: str = None,
        **kwargs
    ):
        self.header_level = header_level if header_level else self.header_level
        self.header_type = header_type if header_type else self.header_type
        self.header_text = header_text if header_text else self.header_text
        self.css_class = css_class if css_class else self.css_class
        self.css_id = css_id if css_id else self.css_id
        self.badge_text = badge_text if badge_text else self.badge_text
        self.badge_class = badge_class if badge_class else self.badge_class
        kwargs['title'] = self.header_text
        super().__init__(**kwargs)

    def get_context_data(self, **kwargs):
        if self.header_type == 'h':
            kwargs['header_class'] = f"h{self.header_level}"
        elif self.header_type == 'display':
            kwargs['header_class'] = f"display-{self.header_level}"

        kwargs['header_level'] = self.header_level
        kwargs['header_text'] = self.header_text
        kwargs['header_type'] = self.header_type
        kwargs['css_class'] = self.css_class
        kwargs['css_id'] = self.css_id
        kwargs['badge_text'] = self.badge_text
        kwargs['badge_class'] = self.badge_class
        return kwargs


class HeaderWithLinkButton(BasicHeader):
    """
    Extends :py:class:`BasicHeader`.

    Header with button that links to something.

    .. deprecated:: 0.13.45
        Use :py:class:`HeaderWithWidget` instead.
    """
    template_name = 'wildewidgets/header_with_link_button.html'

    url: Optional[str] = None
    link_text: Optional[str] = None
    button_class: str = "primary"

    def __init__(
        self,
        url: str = None,
        link_text: str = None,
        button_class: str = None,
        **kwargs
    ):
        warnings.warn('Use wildewidgets.HeaderWithWidget instead', DeprecationWarning, stacklevel=2)
        self.url = url if url else self.url
        self.link_text = link_text if link_text else self.link_text
        self.button_class = button_class if button_class else self.button_class
        super().__init__(**kwargs)

    def get_context_data(self, **kwargs):
        kwargs = super().get_context_data(**kwargs)
        kwargs['url'] = self.url
        kwargs['link_text'] = self.link_text
        kwargs['button_class'] = self.button_class
        return kwargs


class HeaderWithFormButton(BasicHeader):
    """
    Extends :py:class:`BasicHeader`.

    Header with a button that submits a form.

    .. deprecated:: 0.13.45
        Use :py:class:`HeaderWithWidget` instead.
    """
    template_name = 'wildewidgets/header_with_form_button.html'

    url: Optional[str] = None
    button_text: Optional[str] = None

    def __init__(
        self,
        url: str = None,
        button_text: str = None,
        **kwargs
    ):
        warnings.warn('Use wildewidgets.HeaderWithWidget instead', DeprecationWarning, stacklevel=2)
        self.url = url if url else self.url
        self.button_text = button_text if button_text else self.button_text
        super().__init__(**kwargs)

    def get_context_data(self, **kwargs):
        kwargs = super().get_context_data(**kwargs)
        kwargs['url'] = self.url
        kwargs['button_text'] = self.button_text
        return kwargs


class HeaderWithCollapseButton(BasicHeader):
    """
    Extends :py:class:`BasicHeader`.

    .. deprecated:: 0.13.45
        Use :py:class:`HeaderWithWidget` instead.
    """
    template_name: str = 'wildewidgets/header_with_collapse_button.html'

    collapse_id: Optional[str] = None
    button_text: Optional[str] = None
    button_class: Optional[str] = "primary"

    def __init__(
        self,
        collapse_id: str = None,
        button_text: str = None,
        button_class: str = None,
        **kwargs
    ):
        warnings.warn('Use wildewidgets.HeaderWithWidget instead', DeprecationWarning, stacklevel=2)
        self.collapse_id = collapse_id if collapse_id else self.collapse_id
        self.button_text = button_text if button_text else self.button_text
        self.button_class = button_class if button_class else self.button_class
        super().__init__(**kwargs)

    def get_context_data(self, **kwargs):
        kwargs = super().get_context_data(**kwargs)
        kwargs['collapse_id'] = self.collapse_id
        kwargs['button_text'] = self.button_text
        kwargs['button_class'] = self.button_class
        return kwargs


class HeaderWithModalButton(BasicHeader):
    """
    Extends :py:class:`BasicHeader`.

    .. deprecated:: 0.13.45
        Use :py:class:`HeaderWithWidget` instead.
    """
    template_name = 'wildewidgets/header_with_modal_button.html'
    modal_id = None
    button_text = None
    button_class = "primary"

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.modal_id = kwargs.get('modal_id', self.modal_id)
        self.button_text = kwargs.get('button_text', self.button_text)
        self.button_class = kwargs.get('button_class', self.button_class)

    def get_context_data(self, **kwargs):
        kwargs = super().get_context_data(**kwargs)
        kwargs['modal_id'] = self.modal_id
        kwargs['button_text'] = self.button_text
        kwargs['button_class'] = self.button_class
        return kwargs


class HeaderWithWidget(BasicHeader):
    """
    Extends :py:class:`BasicHeader`.

    The base class for more complex headers with buttons or other widgets
    attached. It can be used stand-alone.

    Keyword Args:
        widget: the widget to add to the header.
    """
    template_name = 'wildewidgets/header_with_widget.html'

    def __init__(self, widget: Block = None, **kwargs):
        super().__init__(**kwargs)
        self.widget = widget

    def add_form_button(self, **kwargs):
        """
        Add a form button to the end of the headers. The arguments are those of the :py:class:`wildewidgets.FormButton`.
        """
        self.widget = FormButton(**kwargs)

    def add_link_button(self, **kwargs):
        """
        Add a form button to the end of the headers. The arguments are those of the :py:class:`wildewidgets.LinkButton`.
        """
        self.widget = LinkButton(**kwargs)

    def add_modal_button(self, **kwargs):
        """
        Add a modal button to the end of the headers. The arguments are those of
        the :py:class:`wildewidgets.ModalButton`.
        """
        self.widget = ModalButton(**kwargs)

    def add_collapse_button(self, **kwargs):
        """
        Add a form button to the end of the headers. The arguments are those of
        the :py:class:`wildewidgets.CollapseButton`.
        """
        self.widget = CollapseButton(**kwargs)

    def set_widget(self, widget: Block):
        """
        Add a widget to the header.
        """
        self.widget = widget

    def get_context_data(self, **kwargs):
        kwargs = super().get_context_data(**kwargs)
        kwargs['widget'] = self.widget
        return kwargs


class PageHeader(HeaderWithWidget):
    """
    Extends :py:class:`HeaderWithWidget`.

    Provides a standard page header.
    """
    css_class: str = "my-4"


class CardHeader(HeaderWithWidget):
    """
    Extends :py:class:`HeaderWithWidget`.

    Provides a standard card header.
    """
    css_class: str = "my-3"
    header_level: int = 2


class WidgetListLayoutHeader(HeaderWithWidget):
    """
    Extends :py:class:`HeaderWithWidget`.

    Provides a standard :py:class:`wildewidgets.WidgetListLayout` header.
    """
    css_class: str = "mb-4"
    header_level: int = 2



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from copy import deepcopy
import random
from typing import Any, Dict, List, Optional, Type, Tuple, Union
from urllib.parse import urlencode

from django.core.exceptions import ImproperlyConfigured
from django.core.paginator import InvalidPage, Paginator
from django.forms import Form
from django.db.models import QuerySet, Model
from django.http import Http404

from .base import Block, InputBlock, Widget
from .text import HTMLWidget, LabelBlock
from .buttons import FormButton
from .headers import BasicHeader, CardHeader


class TabbedWidget(Block):
    """
    Extends :py:class:`wildewidgets.widgets.base.Block`.

    This class implements a `Tabler Tabbed Card <https://preview.tabler.io/docs/cards.html>`_.

    Example:

        >>> tab = TabbedWidget()
        >>> tab.add_tab('My First Widget', widget1)
        >>> tab.add_tab('My Second Widget', widget2)
    """

    template_name: str = 'wildewidgets/tab_block.html'
    block: str = 'card'

    slug_suffix: Optional[str] = None
    overflow: str = "auto"

    def __init__(
        self,
        *blocks,
        slug_suffix: str = None,
        overflow: str = None,
        **kwargs
    ):
        self.slug_suffix = slug_suffix if slug_suffix else self.slug_suffix
        self.overflow = overflow if overflow else self.overflow
        super().__init__(*blocks, **kwargs)
        if "style" in self._attributes:
            self._attributes['style'] += f" overflow: {self.overflow};"
        else:
            self._attributes['style'] = f"overflow: {self.overflow};"
        self.tabs = []

    def add_tab(self, name: str, widget: Block) -> None:
        """
        Add a Bootstrap 5 tab named `name` that displays `widget`.

        Args:
            name: the name of the tab
            widget: the block to display in this tab
        """
        self.tabs.append((name, widget))

    def get_context_data(self, **kwargs):
        kwargs['tabs'] = self.tabs
        if not self.slug_suffix:
            self.slug_suffix = random.randrange(0, 10000)
        kwargs['identifier'] = self.slug_suffix
        # kwargs['overflow'] = self.overflow
        return super().get_context_data(**kwargs)


class CardWidget(Block):
    """
    Extends :py:class:`wildewidgets.widgets.base.Block`.

    Renders a `Bootstrap 5 Card <https://getbootstrap.com/docs/5.2/components/card/>`_.

    Keyword Args:
        header: A header widget to display above the card.
        header_text: Text used to build a header widget is one is not provided
        card_title: Card title
        card_subtitle: Card subtitle
        widget: Widget to display in card body
        widget_css: CSS to apply to the widget displayed in the card body
        css_class: CSS to apply to the card itself in addition to the defaults
        overflow:
    """
    template_name: str = 'wildewidgets/card_block.html'

    block: str = 'card'

    header: Optional[BasicHeader] = None
    header_text: Optional[str] = None
    header_css: Optional[str] = None
    card_title: Optional[str] = None
    card_subtitle: Optional[str] = None
    widget: Optional[Widget] = None
    widget_css: Optional[str] = None
    overflow: str = "auto"

    def __init__(
        self,
        *blocks,
        header: BasicHeader = None,
        header_text: str = None,
        header_css: str = None,
        card_title: str = None,
        card_subtitle: str = None,
        widget: Widget = None,
        widget_css: str = None,
        overflow: str = None,
        **kwargs
    ):
        self.header = header if header else deepcopy(self.header)
        self.header_text = header_text if header_text else self.header_text
        self.header_css = header_css if header_css else self.header_css
        if self.header_text and not self.header:
            self.header = CardHeader(header_text=self.header_text)
        self.card_title = card_title if card_title else self.card_title
        self.card_subtitle = card_subtitle if card_subtitle else self.card_subtitle
        self.widget = widget if widget else deepcopy(self.widget)
        self.overflow = overflow if overflow else self.overflow
        self.widget_css = widget_css if widget_css else self.widget_css
        super().__init__(*blocks, **kwargs)
        if "style" in self._attributes:
            self._attributes['style'] += f" overflow: {self.overflow};"
        else:
            self._attributes['style'] = f"overflow: {self.overflow};"

    def get_context_data(self, **kwargs):
        kwargs = super().get_context_data(**kwargs)
        kwargs['header'] = self.header
        kwargs['header_text'] = self.header_text
        kwargs['header_css'] = self.header_css
        kwargs['title'] = self.card_title
        kwargs['subtitle'] = self.card_subtitle
        kwargs['widget'] = self.widget
        kwargs['widget_css'] = self.widget_css
        kwargs['css_class'] = self.css_class

        if not self.widget:
            raise ImproperlyConfigured("You must define a widget.")
        return kwargs

    def set_widget(self, widget, css_class=None):
        """
        Set the widget displayed in the card body

        Args:
            widget (obj): Widget to display in card body
            css_class (str): CSS to apply to the widget displayed in the card body
        """
        self.widget = widget
        self.widget_css = css_class

    def set_header(self, header):
        """
        Set the header widget to display above the card

        Args:
            header (obj): Header widget to display above the card
        """
        self.header = header


class MultipleModelWidget(Block):
    """
    Extends :py:class:`wildewidgets.widgets.base.Block`.

    Base class for :py:class:`PagedModelWidget` and :py:class:`ListModelWidget`.

    Keyword Args:
        model: if this is defined, do ``self.model.objects.all()`` to get our list
            of instances.  Define either this or :py:attr:`queryset`, but not both.
        queryset: use this queryset to get our list of instances.  Define either this or
            :py:attr:`model`, but not both.
        ordering: the ordering to use for the queryset.  This will be applied to both
            :py:attr:`model` and :py:attr:`queryset`
        model_widget: use this widget class to render each model instance.
        model_kwargs: When instantiating :py:attr:`model_widget`, pass this dict into
            the widget constructor as the keyword arguments
        item_label: The label to use for each instance.  This is used in the confirmation
            dialog when deleting instances.

    """
    #: If this is defined, do ``self.model.objects.all()`` to get our list of instances.
    #: Define either this or :py:attr:`queryset`, but not both.
    model: Type[Model] = None
    #: Use this queryset to get our list of instances.  Define either this or
    #: :py:attr:`model`, but not both.
    queryset: QuerySet = None
    #: The ordering to use for the queryset.  This will be applied to both
    #: :py:attr:`model` and :py:attr:`queryset`
    ordering: Union[str, Tuple[str, ...]] = None

    #: Use this widget class to render each model instance.
    model_widget: Type[Widget] = None
    #: When instantiating :py:attr:`model_widget`, pass this dict into the widget constructor
    #: as the keyword arguments
    model_kwargs: Dict[str, Any] = {}

    #: The label to use for each instance.  This is used in the confirmation
    #: dialog when deleting instances.
    item_label: str = ""

    def __init__(
        self,
        *blocks,
        model: Type[Model] = None,
        queryset: QuerySet = None,
        ordering: str = None,
        model_widget: Widget = None,
        model_kwargs: Dict[str, Any] = None,
        item_label: str = None,
        **kwargs
    ) -> None:
        self.model = model if model else self.model
        self.model_widget = model_widget if model_widget else deepcopy(self.model_widget)
        self.model_kwargs = model_kwargs if model_kwargs else deepcopy(self.model_kwargs)
        self.queryset = queryset if queryset else self.queryset
        self.ordering = ordering if ordering else self.ordering
        self.item_label = item_label if item_label else self.item_label
        super().__init__(*blocks, **kwargs)

    def get_item_label(self, instance: Model) -> str:
        """
        Returns the type of the item displayed. This is used in the confirmation
        dialog when deleting.
        """
        return self.item_label

    def get_model_widget(self, object: Model, **kwargs) -> Widget:
        """
        Returns the individual widget to display as part of the list.

        Args:
            object: the model instance to pass into the model specific widget
        """
        # FIXME: we should not be using ``object`` as a variable -- it is a built-in
        if self.model_widget:
            return self.model_widget(object=object, **kwargs)
        else:
            raise ImproperlyConfigured(
                "%(cls)s is missing a model widget. Define "
                "%(cls)s.model_widget or override "
                "%(cls)s.get_model_widget()." % {'cls': self.__class__.__name__}
            )

    def get_model_widgets(self, instances: List[Model]) -> List[Widget]:
        return [
            self.get_model_widget(object=instance, **self.model_kwargs)
            for instance in instances
        ]

    def get_queryset(self) -> Union[List[Model], QuerySet]:
        """
        Return the list of items for this widget.

        The return value must be an iterable and may be an instance of
        :py:class:`QuerySet` in which case query set specific behavior will be
        enabled.
        """
        if self.queryset is not None:
            queryset = self.queryset
            if isinstance(queryset, QuerySet):
                queryset = queryset.all()
        elif self.model is not None:
            # FIXME: why are we using _default_manager here instead of .objects ?
            # The class may have replaced .objects with a more targeted manager and
            # we won't get that if we always hit _default_manager
            queryset = self.model._default_manager.all()
        else:
            raise ImproperlyConfigured(
                "%(cls)s is missing a QuerySet. Define "
                "%(cls)s.model, %(cls)s.queryset, or override "
                "%(cls)s.get_queryset()." % {
                    'cls': self.__class__.__name__
                }
            )
        ordering = self.ordering
        if ordering:
            if isinstance(ordering, str):
                ordering = (ordering,)
            queryset = queryset.order_by(*ordering)
        return queryset


class PagedModelWidget(MultipleModelWidget):
    """
    Extends :py:class:`MultipleModelWidget`.

    A widget that displays a pageable list of widgets
    defined by a queryset. A model or queryset must be provided, and either a model_widget
    or the funtion `get_model_widget` must be provided.

    Args:
        page_kwarg (str, optional): get argument for the page number. Defaults to `page`.
        paginate_by (int, optional): number of widgets per page. Defaults to all widgets.
        extra_url (dict, optional): extra paramters passed in the page links.

    Example::

        PagedModelWidget(
            queryset=project.myobject_set.all(),
            paginate_by=3,
            page_kwarg='myobject_page',
            model_widget=MyObjectWidget,
            model_kwargs={'foo': bar},
            item_label="myobject",
            extra_url={
                'pk': myobject.id,
                '#':'myobjects',
            },
        )
    """
    template_name: str = 'wildewidgets/paged_model_widget.html'
    block: str = "wildewidgets-paged-model-widget"

    page_kwarg: str = 'page'
    paginate_by: int = 25
    max_page_controls: int = 5

    def __init__(
        self,
        *blocks,
        page_kwarg: str = None,
        paginate_by: int = None,
        extra_url: Dict[str, Any] = None,
        **kwargs
    ):
        self.page_kwarg = page_kwarg if page_kwarg else self.page_kwarg
        self.paginate_by = paginate_by if paginate_by else self.paginate_by
        self.extra_url = extra_url if extra_url else {}
        super().__init__(*blocks, **kwargs)

    def get_context_data(self, **kwargs):
        kwargs = super().get_context_data(**kwargs)
        self.request = kwargs.get('request')
        if self.paginate_by:
            paginator = Paginator(self.get_queryset(), self.paginate_by)
            page_number = self.request.GET.get(self.page_kwarg)
            try:
                page_number = int(page_number)
            except (TypeError, ValueError):
                page_number = 1
            try:
                page = paginator.page(page_number)
                kwargs['widget_list'] = self.get_model_widgets(page.object_list)
                kwargs['page_obj'] = page
                kwargs['is_paginated'] = page.has_other_pages()
                kwargs['paginator'] = paginator
                kwargs['page_kwarg'] = self.page_kwarg
                pages = kwargs['page_obj'].paginator.num_pages
                if pages > self.max_page_controls:
                    pages = self.max_page_controls
                page_number = page.number
                max_controls_half = int(self.max_page_controls / 2)
                range_start = 1 if page_number - max_controls_half < 1 else page_number - max_controls_half
                kwargs['page_range'] = range(range_start, range_start + pages)
            except InvalidPage as e:
                raise Http404(
                    'Invalid page (%(page_number)s): %(message)s' % {
                        'page_number': page_number,
                        'message': str(e)
                    }
                )
        else:
            kwargs['widget_list'] = self.get_model_widgets(self.get_queryset().all())
        kwargs['item_label'] = self.item_label
        if self.extra_url:
            anchor = self.extra_url.pop("#", None)
            extra_url = f"&{urlencode(self.extra_url)}"
            if anchor:
                extra_url = f"{extra_url}#{anchor}"
            kwargs['extra_url'] = extra_url
        else:
            kwargs['extra_url'] = ''
        return kwargs


class CrispyFormWidget(Block):
    """
    Extends :py:class:`wildewidgets.widgets.base.Block`.

    A widget that displays a Crispy Form widget.

    Note:

        If the form is present in the template context already, (because your
        view inserted it for you) you don't need to supply the :py:attr:`form`
        keyword argument -- :py:class:`CrispyFormWidget` will pick it up
        automatically.

    Keyword Args:
        form: A crispy form to display. If none is specified,
            it will be assumed that `form` is specified elsewhere in the context.
    """
    template_name: str = 'wildewidgets/crispy_form_widget.html'
    block: str = "wildewidgets-crispy-form-widget"

    def __init__(self, *blocks, form: Form = None, **kwargs):
        super().__init__(*blocks, **kwargs)
        self.form = form

    def get_context_data(self, **kwargs):
        kwargs = super().get_context_data(**kwargs)
        if self.form:
            kwargs['form'] = self.form
        return kwargs


class CollapseWidget(Block):
    """
    Extends :py:class:`wildewidgets.widgets.base.Block`.

    A `Boostrap Collapse widget <https://getbootstrap.com/docs/5.2/components/collapse/>`_.

    It is typically used in conjunction with a :py:class:`CollapseButton`. Pressing the
    :py:class:`CollapseButton` toggles the visibility of the widget.

    Example:

        >>> collapse_id = 'my-collapse-id'
        >>> collapse = CollapseWidget(
            CrispyFormWidget(form=form),
            css_id=collapse_id,
            css_class="pt-3",
        )
        >>> header = CardHeader(header_text="Services")
        >>> header.add_collapse_button(
            text="Manage",
            color="primary",
            target=f"#{collapse_id}",
        )

    Keyword Args:
        css_id: This ID will be shared with the :py:class:`CollapseButton`.

    """
    block: str = "collapse"


class HorizontalLayoutBlock(Block):
    """
    Extends :py:class:`wildewidgets.widgets.base.Block`.

    A container widget intended to display several other widgets aligned horizontally.

    Example:

        >>> layout = HorizontalLayoutBlock(
            LabelBlock(label),
            block1,
            block2,
            justify="end",
            css_class="mt-3",
        )

    Keyword Args:
        align: how to align items veritcally within this widget.  Valid choices: ``start``,
            ``center``, ``end``, ``baselin``, ``stretch``.  See `Bootstrap: Flex,
            justify content <https://getbootstrap.com/docs/5.2/utilities/flex/#align-items>`_.
            If not supplied here and :py:attr:`align` is ``None``, do whatever vertical aligment
            Bootstrap does by default.
        justify: how to align items horizontally within this widget.  Valid choices: ``start``,
            ``center``, ``end``, ``between``, ``around``, ``evenly``.  See `Bootstrap: Flex,
            justify content <https://getbootstrap.com/docs/5.2/utilities/flex/#justify-content>`_.
            If not supplied here and :py:attr:`justify` is ``None``, do whatever horizontal aligment
            Bootstrap does by default.
        flex_size: the Boostrap viewport size below which this will be a vertical list instead
            of a horizontal one.

    """

    #: The valid column content ``justify-content-`` values
    VALID_JUSTIFICATIONS: List[str] = [
        'start',
        'center',
        'end',
        'between',
        'around',
        'evenly'
    ]
    #: The valid column content ``justify-content-`` values
    VALID_ALIGNMENTS: List[str] = [
        'start',
        'center',
        'end',
        'baseline',
        'stretch',
    ]

    #: How to align items veritcally within this widget.  Valid choices: ``start``,
    #: ``center``, ``end``, ``baselin``, ``stretch``.  See `Bootstrap: Flex,
    #: justify content <https://getbootstrap.com/docs/5.2/utilities/flex/#align-items>`_.
    #: If not supplied here and :py:attr:`align` is ``None``, do whatever vertical aligment
    #: Bootstrap does by default.
    align: str = "center"
    #: How to align items horizontally within this widget.  Valid choices: ``start``,
    #: ``center``, ``end``, ``between``, ``around``, ``evenly``.  See `Bootstrap: Flex,
    #: justify content <https://getbootstrap.com/docs/5.2/utilities/flex/#justify-content>`_.
    #: If not supplied here and :py:attr:`justify` is ``None``, do whatever horizontal aligment
    #: Bootstrap does by default.
    justify: str = "between"
    #: the Boostrap viewport size below which this will be a vertical list instead
    #: of a horizontal one.
    flex_size: Optional[str] = None

    def __init__(
        self,
        *blocks,
        align: str = None,
        justify: str = None,
        flex_size: str = None,
        **kwargs
    ):
        self.align = align if align else self.align
        self.justify = justify if justify else self.justify
        self.flex_size = flex_size if flex_size else self.flex_size
        if self.align not in self.VALID_ALIGNMENTS:
            raise ValueError(
                f'"{self.align}" is not a valid vertical alignment value.  Valid values '
                f'are {", ".join(self.VALID_ALIGNMENTS)}'
            )
        if self.justify not in self.VALID_JUSTIFICATIONS:
            raise ValueError(
                f'"{self.justify}" is not a valid horizontal alignment value.  Valid values '
                f'are {", ".join(self.VALID_JUSTIFICATIONS)}'
            )
        super().__init__(*blocks, **kwargs)
        if self.flex_size:
            flex = f"d-{self.flex_size}-flex"
        else:
            flex = "d-flex"
        self.add_class(flex)
        self.add_class(f'align-items-{self.align}')
        self.add_class(f'justify-content-{self.justify}')


class ListModelWidget(MultipleModelWidget):
    """
    Extends :py:class:`MultipleModelWidget`.

    This class provides a list of objects defined by a :py:class:`QuerySet`,
    displayed in a "``list-group``" ``<ul>`` block. By default, a widget will be
    provided that simply displays whatever returns from the conversion of the
    object to a ``str``. If a :py:attr:`remove_url` is provided, an `X` icon to
    the right of each object will act as a button to remove the item.

    Example:

        >>> widget = ListModelWidget(
            queryset=myparent.children,
            item_label='child',
            remove_url=reverse('remove_url') + "?id={}",
        )

    Args:
        remove_url: The url to "POST" to in order to delete or remove the object.

    """
    block: str = 'list-group'
    name: str = 'wildewidgets-list-model-widget'
    tag: str = 'ul'

    #: The url to "POST" to in order to delete or remove the object.
    remove_url: Optional[str] = None

    def __init__(self, *args, remove_url=None, **kwargs):
        self.remove_url = remove_url if remove_url else self.remove_url
        super().__init__(*args, **kwargs)
        widgets = self.get_model_widgets(self.get_queryset().all())
        if not widgets:
            self.add_block(
                Block(
                    f"No {self.item_label}s",
                    tag='li',
                    name='list-group-item',
                    css_class="fw-light fst-italic border"
                )
            )
        for widget in widgets:
            self.add_block(widget)

    def get_remove_url(self, instance: Model) -> str:
        if self.remove_url:
            return self.remove_url.format(instance.id)
        return ""

    def get_confirm_text(self, instance: Model) -> str:
        item_label = self.get_item_label(instance)
        return f"Are you sure you want to remove this {item_label}?"

    def get_object_text(self, instance: Model) -> str:
        return str(instance)

    def get_model_subblock(self, instance: Model):
        if hasattr(instance, 'get_absolute_url'):
            url = instance.get_absolute_url()
            # FIXME: use a Link here
            return Block(HTMLWidget(
                html=f'<a href="{url}"><label>{self.get_object_text(instance)}</label></a>')
            )
        else:
            return Block(self.get_object_text(instance), tag='label')

    def get_model_widget(self, object: Model = None, **kwargs):
        if self.model_widget:
            return super().get_model_widget(object=object, **kwargs)
        widget = HorizontalLayoutBlock(tag='li', name='list-group-item')
        widget.add_block(self.get_model_subblock(object))
        remove_url = self.get_remove_url(object)
        if remove_url:
            widget.add_block(
                FormButton(
                    close=True,
                    action=remove_url,
                    confirm_text=self.get_confirm_text(object),
                ),
            )
        return widget


list_model_card_filter_script = """
var filter_input = document.getElementById("{filter_id}");
filter_input.onkeyup = function(e) {{
    var filter = filter_input.value.toLowerCase();
    document.querySelectorAll("{query} label").forEach(label => {{
        var test_string = label.innerText.toLowerCase();
        if (test_string.includes(filter)) {{
            label.parentElement.classList.remove('d-none');
        }}
        else {{
            label.parentElement.classList.add('d-none');
        }}
    }});
    let children = document.querySelectorAll("{query} li");
    for (let i=0; i < children.length; i++) {{
        let child = children[i];
        child.classList.remove('border-top');
    }};
    for (let i=0; i < children.length; i++) {{
        let child = children[i];
        if (child.classList.contains('d-none')) {{
        }}
        else {{
            child.classList.add('border-top');
            break;
        }}
    }};
}};

"""


class ListModelCardWidget(CardWidget):
    """
    Extends :py:class:`CardWidget`.

    This class provides a :py:class:`CardWidget` with a header and a list of
    objects defined by a :py:class:`QuerySet`, displayed in a
    :py:class:`ListModelWidget`. A filter input is provided to filter the list
    of objects.

    Example:

        >>> widget = ListModelCardWidget(
            queryset = Author.objects.all()[:10]
        )

    Keyword Args:
        list_model_widget_class: The class to use for the list model widget.
            The default is :py:class:`ListModelWidget`.
        list_model_header_class: The class to use for the header.


    """

    list_model_widget_class: Type[Widget] = ListModelWidget
    list_model_header_class: Type[Widget] = None

    def __init__(self, *args, **kwargs):
        self.id_base = f"list_modal_card_{random.randrange(0, 1000)}"
        self.list_model_widget_id = f"{self.id_base}_list_model_widget"
        self.filter_id = f"{self.id_base}_filter"
        self.list_model_widget_class = kwargs.pop("list_model_widget_class", self.list_model_widget_class)
        widget_kwargs = {
            'remove_url': kwargs.pop("remove_url", None),
            'model': kwargs.pop("model", None),
            'model_widget': kwargs.pop("model_widget", None),
            'ordering': kwargs.pop("ordering", None),
            'queryset': kwargs.pop("queryset", None),
            'model_kwargs': kwargs.pop("model_kwargs", {}),
            'item_label': kwargs.pop("item_label", "item"),
            'css_id': self.list_model_widget_id,
        }
        self.widget = self.get_list_model_widget(**widget_kwargs)
        kwargs['widget'] = self.widget
        header_kwargs = {
            'title': kwargs.pop("title", ""),
        }
        kwargs['header'] = self.get_list_model_header(**header_kwargs)
        kwargs['header_css'] = "bg-light"
        filter_label_query = f"#{self.list_model_widget_id}"
        kwargs['script'] = list_model_card_filter_script.format(
            query=filter_label_query,
            filter_id=self.filter_id
        )
        super().__init__(*args, **kwargs)

    def get_list_model_widget(self, *args, **kwargs):
        return self.list_model_widget_class(*args, **kwargs)

    def get_list_model_header(self, *args, **kwargs):
        if self.list_model_header_class:
            return self.list_model_header_class(*args, **kwargs)  # pylint: disable=not-callable
        header = Block(
            Block(
                LabelBlock(
                    f'Filter {self.widget.item_label}s',
                    css_class="d-none",
                    attributes={
                        'for': self.filter_id,
                    },
                ),
                InputBlock(
                    attributes={
                        'type': 'text',
                        'placeholder': f'Filter {self.widget.item_label}s',
                    },
                    css_id=f'{self.id_base}_filter',
                    css_class='form-control',
                ),
                css_class='w-25',
            ),
            Block(""),
            css_class="d-flex flex-row-reverse w-100"
        )
        return header



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from typing import Dict, Any

from .base import Block, TemplateWidget
from .structure import HorizontalLayoutBlock
from .text import CodeWidget


class KeyValueListBlock(Block):

    tag: str = 'ul'
    block: str = "list-group"

    def add_simple_key_value(self, key, value):
        self.add_block(
            HorizontalLayoutBlock(
                Block(key),
                Block(value),
                tag="li",
                css_class="list-group-item",
            )
        )

    def add_code_key_value(self, key, value, language=None):
        self.add_block(
            Block(
                Block(key),
                CodeWidget(
                    code=value,
                    language=language,
                    css_class="m-3",
                ),
                tag="li",
                css_class="list-group-item",
            )
        )


class GravatarWidget(TemplateWidget):

    template_name: str = 'wildewidgets/gravatar.html'
    css_class: str = "rounded-circle"
    size: str = "28"

    def __init__(self, *args, gravatar_url=None, size=None, fullname=None, **kwargs):
        self.gravatar_url = gravatar_url
        self.css_class = kwargs.pop('css_class', self.css_class)
        self.size = size or self.size
        self.fullname = fullname
        super().__init__(*args, **kwargs)

    def get_context_data(self, **kwargs) -> Dict[str, Any]:
        kwargs = super().get_context_data(**kwargs)
        kwargs['gravatar_url'] = self.gravatar_url
        kwargs['css_class'] = self.css_class
        kwargs['size'] = self.size
        kwargs['fullname'] = self.fullname
        return kwargs


class InitialsAvatarWidget(TemplateWidget):

    template_name: str = 'wildewidgets/initials_avatar.html'
    size: int = 28
    color: str = "white"
    background_color: str = "#626976"

    def __init__(self, *args, initials=None, size=None, color=None, background_color=None, fullname=None, **kwargs):
        self.initials = initials
        self.size = size or self.size
        self.color = color or self.color
        self.background_color = background_color or self.background_color
        self.fullname = fullname
        super().__init__(*args, **kwargs)

    def get_context_data(self, **kwargs) -> Dict[str, Any]:
        kwargs = super().get_context_data(**kwargs)
        kwargs['initials'] = self.initials.upper()
        kwargs['fullsize'] = str(self.size)
        kwargs['halfsize'] = str(self.size / 2)
        kwargs['color'] = self.color
        kwargs['background_color'] = self.background_color
        kwargs['fullname'] = self.fullname
        return kwargs



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from collections.abc import Iterable as IterableABC
from copy import deepcopy
from typing import Union, List, Dict, Iterable, Any, Optional

from django.template.loader import get_template
from django.templatetags.static import static


class Widget:
    """
    The base class from which all widgets should inherit.
    """
    title: Optional[Union[str, "Widget"]] = None
    icon: str = 'gear'

    def __init__(
        self,
        title: Union[str, "Widget"] = None,
        icon: str = None,
        **kwargs
    ):
        self.title = title if title else self.title
        self.icon = icon if icon else self.icon
        super().__init__(**kwargs)

    def get_title(self) -> Union[str, "Widget"]:
        return self.title

    def is_visible(self) -> bool:
        return True

    def get_content(self) -> "Widget":
        raise NotImplementedError


class TemplateWidget(Widget):

    #: The name of the template to use to render this widget
    template_name: Optional[str] = None

    def get_content(self, **kwargs) -> str:
        """
        Actually render us to a string.  This is the method that the
        ``wildewidgets`` template tag will call (via
        :py:meth:`wildewidgets.templatetags.wildewidgets.WildewidgetsNode.render`)
        to render the widget into a containing template.

        Returns:
            Our rendered HTML.
        """
        context = self.get_context_data(**kwargs)
        html_template = get_template(self.get_template())
        content = html_template.render(context)
        return content

    def get_template(self) -> str:
        """
        Return the name to the Django template we want to use to render
        this :py:class:`TemplateWidget`.

        Returns:
            The name of the template to use to render us.

        Raises:
            ImproperlyConfigured: no template name was provided.
        """
        return self.template_name

    def get_context_data(self, **kwargs) -> Dict[str, Any]:
        return kwargs

    def __str__(self) -> str:
        return self.get_content()


class Block(TemplateWidget):
    """
    Extends :py:class:`TemplateWidget`.

    Render a single HTML element.

    All the constructor parameters can be set in a subclass of this class as
    class attributes.  Parameters to the constructor override any defined class
    attributes.

    Example:

        >>> block = Block(
            'Hello World',
            tag='a',
            name='foo',
            modifier='bar',
            css_class='blah dah',
            attributes={'href': 'https://example.com'}
        )

        When rendered in the template with the ``wildewdigets`` template tag, this
        will produce::

            <a href="https://example.com" class="foo foo--bar blah dah">Hello World</a>


    Args:
        *blocks: any content to be rendered in the block, including other blocks, or
            plain strings

    Keyword Args:
        tag: the name of the HTML element to use as our tag, e.g. ``div``
        name: This CSS class will be added to the classes to identify this
            block
        modifier: If specified, also add a class named ``{name}--{modifier}`` to the
            CSS classes
        css_class: a string of CSS classes to apply to the block
        css_id: Use this as the ``id`` attribute for the block
        attributes: Set any additional attributes for the block
        data_attributes: Set ``data-bs-`` attributes for the block
        aria_attributes: Set ``aria-`` attributes for the block
        empty: if ``True``, this element uses no close tag
        script: Javascript to attach to this block

    Raises:
        ValueError: empty is ``True``, but contents were added to this block
    """
    template_name: str = "wildewidgets/block.html"

    #: block is the official wildewidgets name of the block; it can't be changed
    #: by constructor kwargs
    block: str = ''

    #: The name of the HTML element to use as our tag, e.g. ``div``
    tag: str = 'div'
    #: The CSS class that will be added to this element to as an identifier for
    #: this kind of block
    name: Optional[str] = None
    #: If specified, also add a class named ``{name}--{modifier}`` to the CSS
    #: classes
    modifier: Optional[str] = None
    #: A string of CSS classes to apply to this block
    css_class: str = ''
    #: The CSS ``id`` for this block
    css_id: Optional[str] = None
    #: A list of strings or Blocks that will be the content for this block
    contents: List[Union[str, Widget]] = []
    #: Additional HTML attributes to set on this block
    attributes: Dict[str, str] = {}
    #: Additional ``data-bs-`` attributes to set on this block
    data_attributes: Dict[str, str] = {}
    #: Additional ``aria-`` attributes to set on this block
    aria_attributes: Dict[str, str] = {}
    #: If ``True``, this block uses no close tag
    empty: bool = False
    #: The Javascript to apply to this block.  It will appear in the HTML directly after
    #: the block appears.  Do not surround this with a ``<script>`` tag; that will be
    #: done for you.
    script: Optional[str] = None

    def __init__(
        self,
        *blocks,
        tag: str = None,
        name: str = None,
        modifier: str = None,
        css_class: str = None,
        css_id: str = None,
        empty: bool = None,
        script: str = None,
        attributes: Dict[str, str] = None,
        data_attributes: Dict[str, str] = None,
        aria_attributes: Dict[str, str] = None
    ):
        super().__init__()
        self._name = name if name is not None else self.name
        self._modifier = modifier if modifier is not None else self.modifier
        self._css_class = css_class if css_class is not None else self.css_class
        if self._css_class is None:
            # Somebody actually went and set our class attribute to ``None``
            self._css_class = ''
        self._css_id = css_id if css_id is not None else self.css_id
        self._tag = tag if tag is not None else self.tag
        self.contents = blocks if blocks else deepcopy(self.contents)
        self.empty = empty if empty is not None else self.empty
        self.script = script if script is not None else self.script
        self._attributes = attributes if attributes is not None else deepcopy(self.attributes)
        self._data_attributes = data_attributes if data_attributes is not None else deepcopy(self.data_attributes)
        self._aria_attributes = aria_attributes if aria_attributes is not None else deepcopy(self.aria_attributes)
        if self.empty and len(blocks) > 0:
            raise ValueError(
                f"{self.__class__.__name__}: This block takes no content because empty=True."
            )
        if self.empty:
            self.template_name = 'wildewidgets/block--simple.html'
        #: The internal list of blocks that are this block's content
        self.blocks: List[Union[str, Widget]] = []
        self.add_blocks()

    @property
    def css_classes(self) -> List[str]:
        """
        Return a list of our CSS classes.

        Note:
            This excludes the set :py:attr:`name`, :py:attr:`block` and
            :py:attr:`modifier` classes.  Read those directly from their
            attributes instead.

        Returns:
            The list of the CSS classes for this block, excluding the
            ``block``, ``name`` and ``modifier`` classes.
        """
        return list(self._css_class.split())

    @css_classes.setter
    def css_classes(self, classes: List[str]) -> None:
        """
        Set our CSS classes to be ``classes``.

        Note:
            Don't try to change :py:attr:`name``, :py:attr:`block`` or
            :py:attr:`modifier`` classes with this; set them by direct assigment
            instead.

        Returns:
            The list of the CSS classes for this block, excluding the
            ``block``, ``name`` and ``modifier`` classes.
        """
        self._css_class = ' '.join(classes)

    def add_class(self, css_class: str) -> None:
        """
        Add a CSS class to our :py:attr:`_css_class` attribute.  This is a
        convenience method to allow us to add a class withouth having to do
        string manipulation.

        Note:
            If you need to change :py:attr:`name` or :py:attr:`modifier`, set them
            directly instead of using this.

        Args:
            css_class: The CSS to add
        """
        # Note: we're purposely not using set() here because Tabler sometimes
        # requires a specific class ordering in order to work properly. set()
        # destroys our ordering.
        if css_class:
            if ' ' in self._css_class:
                # Ensure we don't try to add ``None`` or empty string here
                classes = self._css_class.split()
            else:
                classes = [self._css_class]
            if css_class not in classes:
                classes.append(css_class)
            self._css_class = ' '.join(list(classes))

    def remove_class(self, css_class: str) -> None:
        """
        Remove a CSS class from our :py:attr:`_css_class` attribute.  This is a
        convenience method to allow us to remove a class withouth having to do
        string manipulation.

        Note:
            This method can only remove things from :py:attr:`_css_class`.  You
            can't remove :py:attr:`name`, :py:attr:`modifier` or :py:attr:`block`
            with this method, since those are supposed to be purely informational.

        Args:
            css_class: The CSS to remove
        """
        # Note: we're purposely not using set() here because Tabler sometimes
        # requires a specific class ordering in order to work properly. set()
        # destroys our ordering.
        if css_class:
            # Ensure we don't try to remove ``None`` or empty string here
            classes = self._css_class.split()
            if css_class in classes:
                classes.remove(css_class)
            self._css_class = ' '.join(list(classes))

    def add_blocks(self) -> None:
        """
        Add each block in :py:attr:`contents` to our :py:attr:`blocks` list.

        Note:
            This is here so it can be overridden.
        """
        for block in self.contents:
            self.add_block(block)

    def add_block(self, block: Union[str, Widget]) -> None:
        """
        Add a block to our content.  This will appear inside the tag for this block.

        Args:
            block: the block to add to our content
        """
        if self.empty:
            raise ValueError(
                f"{self.__class__.__name__}: This block takes no content because empty=True."
            )
        self.blocks.append(block)

    def get_script(self) -> Optional[str]:
        """
        Return any javascript to attach to this block.

        Note:
            The Javascript will will appear in the HTML directly after the block
            appears.  Do not surround this with a ``<script>`` tag; that will be
            done for you.

        Returns:
            The fully rendered Javascript for this block, if any.
        """
        return self.script

    def get_context_data(self, **kwargs) -> Dict[str, Any]:
        """
        Update the template context dictionary used when rendering this block.

        Keyword Args:
            **kwargs: the current context dictionary

        Returns:
            The updated context dictionary
        """
        context = super().get_context_data(**kwargs)
        css_class = ' '.join(self.css_classes)
        name = self._name if self._name is not None else ''
        block = self.block if self.block is not None else ''
        if name and self._modifier:
            modifier = f'{name}--{self.modifier}'
        else:
            modifier = ''
        context['tag'] = self._tag
        context['block_name'] = block
        context['name'] = name
        context['css_classes'] = ' '.join([name, modifier, block, css_class]).strip()
        context['script'] = self.get_script()
        context['css_id'] = self._css_id
        context['blocks'] = self.blocks
        context['attributes'] = self._attributes
        context['data_attributes'] = self._data_attributes
        context['aria_attributes'] = self._aria_attributes
        return context


class Container(Block):
    """
    Extends :py:class:`Block`.

    A `Bootstrap container <https://getbootstrap.com/docs/5.2/layout/containers/>`_

    Example::

        >>> c = Container(size='lg')

    Keyword Args:
        size: The max viewport size for the container, in bootstrap sizes, or
            ``fluid``.

    Raises:
        ValueError: ``size`` was not one of the valid sizes
    """
    VALID_SIZES: List[str] = [
        'sm',
        'md',
        'lg',
        'xl',
        'xxl',
        'fluid'
    ]

    #: The max viewport size for the container, in bootstrap sizes, or
    #:``fluid``.
    size: Optional[str] = None

    def __init__(
        self,
        *blocks,
        size: str = None,
        **kwargs
    ):
        self.size = size if size else self.size
        if self.size and self.size not in self.VALID_SIZES:
            raise ValueError(
                f"\"{self.size}\" is not a valid container size.  Valid sizes: {', '.join(self.VALID_SIZES)}")
        super().__init__(*blocks, **kwargs)
        # IMPORTANT: the Tabler rules sometimes expect that the container-* class to
        # be the first class in the class list.  If it is not first, the CSS rules
        # change.  E.g.:
        #
        # .navbar-vertical.navbar-expand-lg > [class^="container"] {
        #     flex-direction: column;
        #     align-items: stretch;
        #     min-height: 100%;
        #     justify-content: flex-start;
        #     padding: 0;
        # }
        classes = self.css_classes
        klass = 'container'
        if self.size:
            klass += f'-{self.size}'
        classes.insert(0, klass)
        self.css_classes = classes


class Link(Block):
    """
    Extends :py:class:`Block`.

    This is a simple ``<a>`` tag.  We made it into its own block because we need `<a>`
    tags so often.

    Example::

        >>> link = Link('click here', url='https://example.com')

    Keyword Args:
        contents: The contents of the ``<a>``.  This can be string, a
            :py:class:`Block`, or an iterable of strings and :py:class:`Block`
            objects.
        url: The URL of the ``<a>``
        title: The value of the ``title`` tag for the link
        role: The value of the ``role`` attribute for the link
        target: The target for this link, aka the context in which the linked
            resource will open.
    """

    tag: str = 'a'

    #: The contents of the ``<a>``.  This can be string, a :py:class:`Block`,
    #: or an iterable of strings and :py:class:`Block` objects.
    contents: Union[str, Block, Iterable[Union[str, Block]]] = None
    #: The URL of the ``<a>``
    url: str = '#'
    #: The value of the ``role`` attribute for the link
    role: str = 'link'
    #: The value of the ``title`` attribute for the link
    title: Optional[str] = None
    #: The target for this link, aka the context in which the linked resource
    #: will open.
    target: Optional[str] = None

    def __init__(
        self,
        *contents: Union[str, Block],
        url: str = None,
        role: str = None,
        title: str = None,
        target: str = None,
        **kwargs
    ):
        self.url = url if url else self.url
        self.role = role if role else self.role
        self.title = title if title else self.title
        self.target = target if target else self.target
        # FIXME: validate url
        # FIXME: validate that title and role are htmleescaped
        if contents:
            self.contents: Iterable[Union[str, Block]] = contents
        else:
            c = self.contents
            if isinstance(c, str):
                self.contents = [c]
            elif isinstance(c, Block):
                self.contents = [deepcopy(c)]
            elif isinstance(c, IterableABC):
                self.contents = deepcopy(c)
            else:
                self.contents = []
        super().__init__(*self.contents, **kwargs)
        self._attributes['href'] = self.url
        self._attributes['role'] = self.role
        if self.title:
            self._attributes['title'] = self.title
        if self.target:
            self._attributes['target'] = self.title


class UnorderedList(Block):
    """
    Extends :py:class:`Block`.

    An HTML unordered list, aka a ``<ul>`` tag.

    This wraps each item in :py:attr:`contents` with ``<li>``.

    Example:

        >>> ul = UnorderedList('foo', Link('bar', url='https://example.com'))
    """

    tag: str = 'ul'

    def add_block(self, block: TemplateWidget) -> None:
        """
        Wrap ``block``  in an ``<li>`` and append it to :py:attr:`blocks`.
        """
        self.blocks.append(Block(block, tag='li'))


class OrderedList(Block):
    """
    Extends :py:class:`UnorderedList`.

    An HTML ordered list, aka a ``<ol>`` tag.

    This wraps each item in :py:attr:`contents` with ``<li>``.

    Example:

        >>> ul = OrderedList('foo', Link('bar', url='https://example.com'))
    """

    tag: str = 'ol'


# Backwards compatibility

class HTMLList(UnorderedList):

    def __init__(self, *args, items: List[str] = None, **kwargs):
        if not items:
            items = []
        super().__init__(*args, **kwargs)
        self.contents = items if items else self.contents


class Image(Block):
    """
    An ``<img>``::

        <img src="image.png" alt="My Image" width="100%">

    Example:

        >>> block = Image(src=static('myapp/images/img.png'), alt='The Image')

    Keyword Args:
        src: the URL of the image.  Typically this will be something like
            ``static('myapp/images/image.png')`` The default is to use a
            placeholder image to remind you that you need to fix this.
        height: the value of the ``height`` attribute for the ``<img>``
        width: the value of the ``width`` attribute for the ``<img>``
        alt: the value of the ``alt`` tag for the ``<img>``. If this is not
            set either here or as a class attribute, we'll raise ``ValueError`` to
            enforce WCAG 2.0 compliance.

    Raises:
        ValueError: no ``alt`` was provided
    """

    tag: str = 'img'
    block: str = 'image'

    #: The URL of the image.  Typically this will be something like :
    #``static('myapp/images/image.png')``.  The default is to use a placeholder
    #: image to remind you that you need to fix this.
    src: Optional[str] = None
    #: the value of the ``width`` attribute for the <img>
    width: Optional[str] = None
    #: the value of the ``height`` attribute for the <img>
    height: Optional[str] = None
    #: The value of the ``alt`` tag for the <img>.  If this is not set either
    #: here or in our contructor kwargs, we'll raise ``ValueError`` (to enforce
    #: ADA)
    alt: Optional[str] = None

    def __init__(
        self,
        src: str = None,
        height: str = None,
        width: str = None,
        alt: str = None,
        **kwargs
    ):
        self.src = src if src else self.src
        if not self.src:
            self.src = static('wildewidgets/images/placeholder.png')
        # FIXME: validate src as a URL/Path
        self.width = width if width else self.width
        self.height = height if height else self.height
        self.alt = alt if alt else self.alt
        super().__init__(**kwargs)
        if self.src:
            self._attributes['src'] = src
        if self.width:
            self._attributes['width'] = width
        if not self.alt:
            raise ValueError('You must provide an "alt" attribute for your Image')
        self._attributes['alt'] = self.alt


class LinkedImage(Link):
    """
    Extends :py:class:`Link`.

    An ``<img>`` wrapped in an ``<a>``::

        <a href="#">
            <img src="image.png" alt="My Image" width="100%">
        </a>

    Note:

        If you want to modify the encapsulated image (to add css classes, for
        example), you can do so by modifying the attributes on :py:attr:`image`
        after constructing the ``LinkedImage``::

            >>> block = LinkedImage(image_src='image.png', image_alt='My Image',
                url='http://example.com')
            >>> block.image.add_class('my-extra-class')
            >>> block.image._css_id = 'the-image'

    Keyword Args:
        image_src: the URL of the image.  Typically this will be something like
            ``static('myapp/images/image.png')``
        image_width: the value of the ``width`` attribute for the ``<img>``
        image_alt: the value of the ``alt`` tag for the ``<img>``. If this is not
            set either here or as a class attribute, we'll raise ``ValueError`` to
            enforce WCAG 2.0 compliance.

    Raises:
        ValueError: no ``alt`` was provided
    """

    block: str = 'linked-image'

    #: The URL of the image.  Typically this will be something like
    #: ``static('myapp/images/image.png')``
    image_src: Optional[str] = None
    #: the value of the ``width`` attribute for the ``<img>``.
    image_width: Optional[str] = None
    #: The value of the ``alt`` tag for the ``<img>``.  If this is not set either
    #: here or in our contructor kwargs, we'll raise ``ValueError`` to enforce
    #: WCAG 2.0 compliance.
    image_alt: Optional[str] = None

    def __init__(
        self,
        image_src: str = None,
        image_width: str = None,
        image_alt: str = None,
        **kwargs
    ):
        self.image_src = image_src if image_src else self.image_src
        if not self.image_src:
            self.image_src = static('sphinx_hosting/images/placeholder.png')
        self.image_width = image_width if image_width else self.image_width
        self.image_alt = image_alt if image_alt else self.image_alt
        #: The actual image block that we will wrap with an ``<a>``
        self.image: Image = Image(src=self.image_src, width=self.image_width, alt=self.image_alt)
        super().__init__(self.image, **kwargs)


class WidgetStream(Block):
    template_name: str = 'wildewidgets/widget_stream.html'
    block: str = 'widget-stream'
    widgets: List[Widget] = []

    def __init__(self, widgets: List[Widget] = None, **kwargs):
        self._widgets = widgets if widgets else deepcopy(self.widgets)
        super().__init__(**kwargs)

    @property
    def is_empty(self) -> bool:
        return len(self._widgets) == 0

    def add_widget(self, widget, **kwargs):
        wrapper = Block(widget, **kwargs)
        wrapper.block = f"{self.block}__widget"
        self._widgets.append(wrapper)

    def get_context_data(self, **kwargs) -> Dict[str, Any]:
        context = super().get_context_data(**kwargs)
        context['widgets'] = self._widgets
        return context


class InputBlock(Block):
    """
    Extends :py:class:`Block`.

    A block for rendering a ``<input>`` element.
    """
    empty: bool = True
    tag: str = 'input'



from setuptools import setup, find_packages

with open("README.md", "r", encoding='utf-8') as fh:
    long_description = fh.read()

setup(
    name='django-wildewidgets',
    version="0.14.0",
    packages=find_packages(exclude=['bin']),
    include_package_data=True,
    description='django-wildewidgets is a Django design library providing several tools for '
                'building full-featured, widget-based web applications with a standard, '
                'consistent design, based on Bootstrap.',
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Caltech IMSS ADS",
    author_email="imss-ads-staff@caltech.edu",
    url='https://github.com/caltechads/django-wildewidgets',
    keywords=['design', 'widget', 'django'],
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        "Framework :: Django",
        "Framework :: Django :: 3.2",
        "Framework :: Django :: 4.0",
        "Framework :: Django :: 4.1",
        "Topic :: Software Development :: User Interfaces",
    ],
)
